# -*- coding: utf-8 -*- vim: ts=8 sts=4 sw=4 si et tw=79
"""\
dxtools - Dexterity tools

Dexterity should be installed to use this module;
otherwise, some functions might fail
"""

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"

# Local imports:
from ._dx4 import get_first_text_as_html
from ._dx5 import generate_all_texts, get_all_texts
